#include <stdlib.h>
#include "treemap.h"

typedef struct node node_t;

struct treemap
{
  node_t *root;
};

struct node
{
  int key;
  void *value;
  node_t *left;
  node_t *right;
};

static node_t *node_new(int key, void *value)
{
  node_t *n = calloc(1, sizeof(node_t));
  n->key = key;
  n->value = value;
  return n;
}

/// Helper function for treemap_insert:
static node_t *node_insert(node_t *n, int key, void *value)
{
  /// Hint -- use this as a helper function for treemap_insert().
  /// Implement it as a recursive function that either:
  /// 1) inserts a new node with the demanded key and value pointer, or
  /// 2) if the key already exists: replaces the value pointer, or
  /// 3) continues the search in the subtrees.
  if (n == NULL)
    {                     // If no such key, insert a new node here:
      n = node_new(key, value);
    }
  else if (n->key == key) // Else check if the key already exists ...
    {
      n->value = value;   // ... if it does, replace the value pointer ...
    }
  else if (n->key > key)  // ... if not, try inserting into the left subtree ...
    {
      n->left = node_insert(n->left, key, value);
    }
  else if (n->key < key)  // ... -- OR -- into the right subtree if key is BIG.
    {
      n->right = node_insert(n->right, key, value);
    }
  return n;
}

/// Insert value v under key k in the tree -- if there is
/// already a value v' with key k, return v' and replace
/// v' with v in the tree.
void *treemap_insert(treemap_t *t, int key, void *value)
{
  /// TODO: lägg till enligt spec
  /// Remember there are instructions in the exam text
  /// for how to implement insert
  /// Hint: use a recursive help function (or double pointers)
  //t->root = node_new(key, value);
  t->root = node_insert(t->root, key, value);
  return NULL;
}

/// Helper function for treemap_lookup:
static node_t *node_lookup(node_t *n, int key, void** adr_of_value)
{
  /// Hint -- use this as a helper function for treemap_lookup().
  /// Implement it as a recursive function that:
  /// 1) if the key is found: updates the value pointer, or
  /// 2) continues the search in the subtrees.
  if (n != NULL)
    {
      if (n->key == key)            // Check if the key is in this node,
        {
          *adr_of_value = n->value; // if it is, update the value pointer,
        }
      else if (n->key > key)        // if not, try searching the left subtree,
        {
          n->left = node_lookup(n->left, key, adr_of_value);
        }
      else if (n->key < key)        // or the right subtree.
        {
          n->right = node_lookup(n->right, key, adr_of_value);
        }
    }
  return n;
}

/// Return the value associated with key k in the tree.
/// If key k is not in the tree, return NULL. 
void *treemap_lookup(treemap_t *t, int key)
{
  /// TODO: lägg till enligt spec
  void* result = NULL;
  t->root = node_lookup(t->root, key, &result);
  return result;
}

treemap_t *treemap_create()
{
  return calloc(1, sizeof(treemap_t)); 
}

/// Helper function for treemap_destroy
static void node_destroy(node_t *n)
{
  /// TODO: frivillig hjälpfunktion
  /// Hint -- use this as a helper function to treemap_destroy
  /// Implement it as a recursive function that first visits
  /// the subtrees and then destroys n
  /// Postorder traversal
  if (n == NULL) return;
  /// Recurse into left subtree
  if (n->left != NULL) node_destroy(n->left);
  /// Recurse into right subtree
  if (n->right != NULL) node_destroy(n->right);
  /// Free the memory of the current node n:
  free(n);
}

/// Removes the treemap from memory, but does not call
/// free on any value pointer passed to it.
void treemap_destroy(treemap_t *t)
{
  /// TODO: lägg till enligt spec
  //free(t->root);
  node_destroy(t->root);
  free(t);
}

/// Hint: look at this function for inspiration
bool treemap_has_key(treemap_t *t, int key)
{
  node_t *n = t->root;

  while (n)
    {
      if (n->key == key) return true;

      if (n->key > key)
        {
          n = n->left;
        }
      else
        {
          n = n->right;
        }
    }

  /// If we came here without finding it, it can't exist
  return false;
}

/// ===================================================
/// From here on: given functions used by driver 
/// ===================================================

// Help function to treemap_keys
// You do not need to read or change this function
static void node_keys(node_t *n, int **keys)
{
  if (n == NULL) return;

  **keys = n->key;
  ++*keys;
  node_keys(n->left, keys);
  node_keys(n->right, keys);
}

// You do not need to read or look at this function
int *treemap_keys(treemap_t *t)
{
  int *keys = calloc(treemap_size(t) + 1, sizeof(int));
  int *cursor = keys;
  node_keys(t->root, &cursor);

  return keys;
}

/// Helper function for treemap_size
static void node_size(node_t *n, int *acc)
{
  /// Postorder traversal
  if (n == NULL) return;
  /// Recurse into left subtree
  if (n->left != NULL) node_size(n->left, acc);
  /// Recurse into right subtree
  if (n->right != NULL) node_size(n->right, acc);

  /// Increment the accumulator
  *acc += 1;
}

int treemap_size(treemap_t *t)
{
  int result = 0;
  node_size(t->root, &result);
  return result;
}
